unit RegExp;

interface

{$IFDEF StonyBrook}
  uses Strings;
{$ENDIF}

const
  RegExp_NSUBEXP = 128;
  MAX_NSUBEXP    = RegExp_NSUBEXP - 1;

type
{$IFNDEF StonyBrook}
  CharPtr = PChar;
{$ENDIF}
  SubExpArrPtr = ^SubExpArr;
  SubExpArr = array[0..MAX_NSUBEXP] of CharPtr;

  RegExprPtr = ^RegExpr;
  RegExpr = record
    startp, endp : SubExpArr;
    regstart, reganch : char; { Internal use only. }
    regmust : CharPtr;          { Internal use only. }
    regmlen : word;           { Internal use only. }
    rprogram : array[0..0] of byte;  { Unwarranted chumminess with compiler. }
  end;

  StringTablePtr = ^StringTable;
  StringTable = array[0..0] of CharPtr;

  RegularExpressionPtr = ^RegularExpression;
  RegularExpression = object
  private
    regparse    : CharPtr;          { Input-scan pointer. }
    regreplace  : CharPtr;          { Mask for replace }
    regcase     : boolean;        { case sensitivity }
    regnpar     : integer;        { () count. }
    regdummy    : char;
    regcode     : CharPtr;          { Code-emit pointer; &regdummy = don't. }
    regsize     : longint;        { Code size. }
    regErrorMsg : ^string;
    reginput    : CharPtr;          { String-input pointer. }
    regbol      : CharPtr;          { Beginning of input, for ^ check. }
    regstartp   : StringTablePtr; { Pointer to startp array. }
    regendp     : StringTablePtr; { Ditto for endp. }
    mask        : RegExprPtr;

    procedure regerror( const ErrStr : string );
    function regcomp( exp : CharPtr ) : RegExprPtr;
    function regexec( prog : RegExprPtr; st : CharPtr ) : boolean;
    function reg( paren : word; var flagp : word ) : CharPtr;
    function regbranch( var flagp : word ) : CharPtr;
    function regpiece( var flagp : word ) : CharPtr;
    function regatom( var flagp : word ) : CharPtr;
    function regnode( op : char ) : CharPtr;
    function regnext( p : CharPtr ) : CharPtr;
    procedure reginsert( op : char; opnd : CharPtr );
    procedure regtail( p, val : CharPtr );
    procedure regoptail( p, val : CharPtr );
    function regtry( prog : RegExprPtr; st : CharPtr ) : boolean;
    function regmatch( prog : CharPtr ) : boolean;
    function regrepeat( p : CharPtr ) : word;
    procedure regsub( prog : RegExprPtr; source, dest : CharPtr );
    procedure regc( b : char); { emit (if appropriate) a byte of code }
  {$IFOPT D+}
    procedure regdump( r : RegExprPtr );
    function regprop( op : CharPtr ) : string;
  {$ENDIF}
  public
    constructor Init( MaskStr : CharPtr );
    destructor Done; virtual;
    procedure setReplaceMask( st : CharPtr );
    function isValid : boolean;
    function maskMatch( addr : CharPtr; caseSense : boolean ) : boolean;
    function maskReplace( addr : CharPtr; caseSense : boolean ) : boolean;
    function getRegErrorMsg : string;
  end;


implementation

uses Strings, OpAsciiz,
{$IFDEF StonyBrook}
  OpInline,
{$ENDIF}
  OpString;
{$R-}

type
  RegOpPtr = ^RegOp;
  RegOp = packed record
    OP : byte;
    NEXT : word;
  end;

const
  OPHDR = SizeOf(RegOp);

  MAGIC = 0234;

  { definition number  opnd? meaning }
  PEND    = 0;   { no End of program. }
  BOL     = 1;   { no Match "" at beginning of line. }
  EOL     = 2;   { no Match "" at end of line. }
  ANY     = 3;   { no Match any one character. }
  ANYOF   = 4;   { str  Match any character in this string. }
  ANYBUT  = 5;   { str  Match any character not in this string. }
  BRANCH  = 6;   { node Match this alternative, or the next... }
  BACK    = 7;   { no Match "", "next" ptr points backward. }
  EXACTLY = 8;   { str  Match this string. }
  NOTHING = 9;   { no Match empty string. }
  STAR    = 10;  { node Match this (simple) thing 0 or more times. }
  PLUS    = 11;  { node Match this (simple) thing 1 or more times. }
  OPEN    = 20;  { no Mark this point in input as start of #n. }
                 {  OPEN+1 is number 1, etc. }
  CLOSE   = 30;  { no Analogous to OPEN. }

{
 * Opcode notes:
 *
 * BRANCH	The set of branches constituting a single choice are hooked
 *		together with their "next" pointers, since precedence prevents
 *		anything being concatenated to any individual branch.  The
 *		"next" pointer of the last BRANCH in a choice points to the
 *		thing following the whole choice.  This is also where the
 *		final "next" pointer of each individual branch points; each
 *		branch starts with the operand node of a BRANCH node.
 *
 * BACK		Normal "next" pointers all implicitly point forward; BACK
 *		exists to make loop structures possible.
 *
 * STAR,PLUS	'?', and complex '*' and '+', are implemented as circular
 *		BRANCH structures using BACK.  Simple cases (one character
 *		per match) are implemented with STAR and PLUS for speed
 *		and to minimize recursive plunges.
 *
 * OPEN,CLOSE	...are numbered at compile time.
}

{
 * Flags to be passed up and down.
}
  HASWIDTH = 01;  { Known never to match null string. }
  SIMPLE   = 02;  { Simple enough to be STAR/PLUS operand. }
  SPSTART  = 04;  { Starts with * or +. }
  WORST    = 0;   { Worst case. }

  META : array[0..11] of char = ('^', '$', '.', '[', '(', ')',
                                 '|', '?', '+', '*', '\', Chr(0));

var
  LastRegErrorMsg : string;

{$IFDEF StonyBrook}
function OPERAND( p : CharPtr ) : CharPtr [INLINE];
  begin
    OPERAND := CharPtr(Ptr(Seg(p^),Ofs(p^) + OPHDR));
  end;
{$ENDIF}

{$IFDEF StonyBrook}
function ISMULT( c : char ) : boolean [INLINE];
{$ELSE}
function ISMULT( c : char ) : boolean;
{$ENDIF}
  begin
    if (c = '*') or (c = '+') or (c = '?') then ISMULT := true
    else ISMULT := false;
  end;

function strcspn(s1, s2 : CharPtr) : word;
  var
    scan1, scan2 : CharPtr;
    count : word;
  begin
    count := 0;
    scan1 := s1;
    while scan1^ <> #0 do
      begin
        scan2 := s2;
        while scan2^ <> #0 do
          begin
            if scan1^ = scan2^ then
              begin
                strcspn := count; exit;
              end;
            Inc(scan2);
          end;
        Inc(count);
        Inc(scan1);
      end;
    strcspn := count;
  end;

{ ----------   convert escape sequences  -------------}
const
  NOT_HEX = 16;
  ET_END  = 7;
type
  ET = record
  i, o : char;
  end;
const
  hex_val : array[0..(Ord('f')-Ord('A'))] of byte = (
            10,11,12,13,14,15, 0, 0,
             0, 0, 0, 0, 0, 0, 0, 0,
             0, 0, 0, 0, 0, 0, 0, 0,
             0, 0, 0, 0, 0, 0, 0, 0,
            10,11,12,13,14,15 ) ;
  escape_test : array[0..ET_END] of ET = (
  (i:'n';o:#10), (i:'t';o:#9), (i:'f';o:#12), (i:'b';o:#8),
  (i:'r';o:#13), (i:'a';o:#7), (i:'v';o:#11), (i:#0;o:#0) );

function escape( var p : CharPtr) : char;
  function ctohex( c : char ) : byte;
  { interpret 1 character as hex }
    var t : byte;
    begin
      if c in ['0'..'9'] then
        begin
          ctohex := Ord(c) - Ord('0');
          exit;
        end;
      if c in ['A'..'f'] then
        begin
          t := hex_val[Ord(c) - Ord('A')];
          if t <> 0 then
            begin
              ctohex := t; exit;
            end;
        end;
      ctohex := NOT_HEX;
    end;

  var
    x, xx, i : word;
  begin
    if p^ = Chr(0) then
      begin
        escape := '\'; exit;
      end;

    escape_test[ET_END].i := p^;
    i := 0;
    while escape_test[i].i <> p^  do Inc(i);
    if i <> ET_END then { in table }
      begin
        Inc(p);
        escape := escape_test[i].o;
        exit;
      end;

    if p^ in ['0'..'7'] then
      begin
        x := Ord(p^) - Ord('0');
        Inc(p);
        if p^ in ['0'..'7'] then
          begin
            x := (x shl 3) + (Ord(p^) - Ord('0'));
            Inc(p);
            if p^ in ['0'..'7'] then
              begin
                x := (x shl 3) + (Ord(p^) - Ord('0'));
                Inc(p);
              end;
          end;
        escape := Chr(x);
        exit;
      end;

    if p^ = 'x' then { might be a hex digit }
      begin
        Inc(p);
        x := ctohex(p^);
        if x = NOT_HEX then
          begin
            escape := 'x'; exit;
          end;
        { look for another hex digit }
        Inc(p);
        xx := ctohex(p^);
        if xx <> NOT_HEX then
          begin
            x := (x shl 4) + xx;
            Inc(p);
          end;
        escape := Chr(x);
        exit;
      end;
    { anything else \c -> c }
    escape := p^;
    Inc(p);
  end;

type
  SearchFunc = function(Str: CharPtr; C: Char): CharPtr;
  CompareFunc = function(Str1, Str2: CharPtr; MaxLen: Word): Integer;
var
  FindChar : SearchFunc;
  CompString : CompareFunc;

function StrIScan(Str: CharPtr; C: Char): CharPtr; far;
  var
    c1, c2 : char;
    p1, p2 : CharPtr;
  begin
    c1 := UpCase(C); c2 := LoCase(C);
    if c1 = c2 then StrIScan := StrScan(Str, C)
    else
      begin
        p1 := StrScan(Str, c1); p2 := StrScan(Str, c2);
        if p1 <> nil then
          begin
            StrIScan := p1;
            if p2 <> nil then
            {$IFDEF StonyBrook}
              if PtrToLong(p2) < PtrToLong(p1) then StrIScan := p2;
            {$ELSE}
              if p2 < p1 then StrIScan := p2;
            {$ENDIF}
          end
        else StrIScan := p2;
      end;
  end;

{
 - regcomp - compile a regular expression into internal code
 *
 * We can't allocate space until we know how big the compiled form will be,
 * but we can't compile it (and thus know how big it is) until we've got a
 * place to put the code.  So we cheat:  we compile it twice, once with code
 * generation turned off and size counting turned on, and once "for real".
 * This also means that we don't allocate space until we are sure that the
 * thing really will compile successfully, and we never have to move the
 * code and thus invalidate pointers into it.  (Note that it has to be in
 * one piece because free() must be able to free it all.)
 *
 * Beware that the optimization-preparation code in here knows about some
 * of the structure of the compiled regexp.
}
function RegularExpression.regcomp( exp : CharPtr ) : RegExprPtr;
  var
    r : RegExprPtr;
    scan, longest : CharPtr;
    len, flags : word;
  begin
    if exp = nil then
      begin
        regerror('NULL argument');
        regcomp := nil; exit;
      end;

    { First pass: determine size, legality. }
    regparse := exp;
    regnpar := 1;
    regsize := 0;
    regcode := @regdummy;
    regc(Chr(MAGIC));
    if reg(0, flags) = nil then
      begin
        regcomp := nil; exit;
      end;

    { Small enough for pointer-storage convention? }
    if regsize >= 32767 then   { Probably could be 65535L. }
      begin
        regerror('regexp too big');
        regcomp := nil; exit;
      end;

    { Allocate space. }
    GetMem(r, SizeOf(RegExpr) + Word(regsize));
    if r = nil then
      begin
        regerror('out of space');
        regcomp := nil; exit;
      end;

    { Second pass: emit code. }
    regparse := exp;
    regnpar := 1;
    regcode := @r^.rprogram;
    regc(Chr(MAGIC));
    if reg(0, flags) = nil then
      begin
        regcomp := nil; exit;
      end;

    { Dig out information for optimizations. }
    r^.regstart := #0; { Worst-case defaults. }
    r^.reganch := #0;
    r^.regmust := nil;
    r^.regmlen := 0;
    scan := CharPtr(@r^.rprogram); Inc(scan);     { First BRANCH. }
    if RegOpPtr(regnext(scan))^.OP = PEND then  { Only one top-level choice. }
      begin
        Inc(scan,OPHDR);

        { Starting-point info. }
        if RegOpPtr(scan)^.OP = EXACTLY then
	{$IFDEF StonyBrook}
          r^.regstart := OPERAND(scan)^
	{$ELSE}
          r^.regstart := (scan+OPHDR)^
	{$ENDIF}
        else if RegOpPtr(scan)^.OP = BOL then
          Inc(r^.reganch);

        {
         * If there's something expensive in the r.e., find the
         * longest literal string that must appear and make it the
         * regmust.  Resolve ties in favor of later strings, since
         * the regstart check works with the beginning of the r.e.
         * and avoiding duplication strengthens checking.  Not a
         * strong reason, but sufficient in the absence of others.
        }
        if (flags and SPSTART) <> 0 then
          begin
          longest := nil;
          len := 0;
          while scan <> nil do
            begin
              if (RegOpPtr(scan)^.OP = EXACTLY) and
	      {$IFDEF StonyBrook}
                 (StrLen(OPERAND(scan)) >= len) then
	      {$ELSE}
                 (StrLen(scan+OPHDR) >= len) then
	      {$ENDIF}
                begin
		{$IFDEF StonyBrook}
                  longest := OPERAND(scan);
                  len := StrLen(OPERAND(scan));
		{$ELSE}
                  longest := scan+OPHDR;
                  len := StrLen(scan+OPHDR);
		{$ENDIF}
                end;
              scan := regnext(scan);
            end;
          r^.regmust := longest;
          r^.regmlen := len;
          end;
    end;

  {$IFOPT D+}
    regdump(r);
  {$ENDIF}
    regcomp := r;
  end;

procedure RegularExpression.regc( b : char); { emit (if appropriate) a byte of code }
  begin
    if regcode <> @regdummy then
      begin
        regcode^ := b;
        Inc(regcode);
      end
    else
      Inc(regsize);
  end;

{
 - reg - regular expression, i.e. main body or parenthesized thing
 *
 * Caller must absorb opening parenthesis.
 *
 * Combining parenthesis handling with the base level of regular expression
 * is a trifle forced, but the need to tie the tails of the branches to what
 * follows makes it hard to avoid.
}
function RegularExpression.reg( paren : word; { Parenthesized? }
                                var flagp : word) : CharPtr;
  var
    ret, br, ender : CharPtr;
    parno, flags : word;
  begin
    flagp := HASWIDTH;  { Tentatively. }

    { Make an OPEN node, if parenthesized. }
    if paren <> 0 then
      begin
        if regnpar >= RegExp_NSUBEXP then
          begin
            regerror('too many ()');
            reg := nil; exit;
          end;
        parno := regnpar;
        Inc(regnpar);
        ret := regnode( Chr(OPEN+parno) );
      end
    else
      ret := nil;

    { Pick up the branches, linking them together. }
    br := regbranch(flags);
    if br = nil then
      begin
        reg := nil; exit;
      end;
    if ret <> nil then
      regtail(ret, br) { OPEN -> first. }
    else
      ret := br;
    if (flags and HASWIDTH) = 0 then
      flagp := flagp and (not HASWIDTH);
    flagp := flagp or (flags and SPSTART);
    while ( regparse^ = '|') do
      begin
        Inc(regparse);
        br := regbranch(flags);
        if br = nil then
          begin
            reg := nil; exit;
          end;
        regtail(ret, br); { BRANCH -> BRANCH. }
        if (flags and HASWIDTH) = 0 then
          flagp := flagp and (not HASWIDTH);
        flagp := flagp or (flags and SPSTART);
      end;

    { Make a closing node, and hook it on the end. }
    if paren <> 0 then
      ender := regnode( Chr(CLOSE + parno) )
    else
      ender := regnode( Chr(PEND) );
    regtail(ret, ender);

    { Hook the tails of the branches to the closing node. }
    br := ret;
    while br <> nil do
      begin
        regoptail(br, ender);
        br := regnext(br);
      end;

    { Check for proper termination. }
    if (paren <> 0) and  (regparse^ <> ')') then
      begin
        Inc(regparse);
        regerror('unmatched ()');
        reg := nil; exit;
      end
    else
      begin
        if paren <> 0 then Inc(regparse);
        if (paren = 0) and (regparse^ <> #0) then
          begin
            if regparse^ = ')' then
              begin
                regerror('unmatched ()');
                reg := nil; exit;
              end
            else
              begin
                regerror('junk on end');
                reg := nil; exit;
              end;  { "Can't happen". }
            { NOTREACHED }
          end;
      end;

    reg := ret;
  end;

constructor RegularExpression.Init( MaskStr : CharPtr );
  begin
    regparse := nil;
    regreplace := nil;
    regnpar := 0;
    regdummy := Chr(0);
    regcode := nil;
    regsize := 0;
    reginput := nil;
    regbol := nil;
    regstartp := nil;
    regendp := nil;
    regErrorMsg := nil;
    mask := regcomp(MaskStr);
  end;

destructor RegularExpression.Done;
  begin
    if mask <> nil then
      FreeMem(mask, SizeOf(RegExpr) + Word(regsize));
    if regreplace <> nil then StrDispose( regreplace );
  end;

procedure RegularExpression.regerror( const ErrStr : string );
  begin
    regErrorMsg := @ErrStr;
    LastRegErrorMsg := ErrStr;
  end;

{
 - regbranch - one alternative of an | operator
 *
 * Implements the concatenation operator.
}
function RegularExpression.regbranch( var flagp : word ) : CharPtr;
  var
    ret, chain, latest : CharPtr;
    flags : word;
  begin

    flagp := WORST;   { Tentatively. }

    ret := regnode(Chr(BRANCH));
    chain := nil;
    while (regparse^ <> #0) and (regparse^ <> '|') and (regparse^ <> ')') do
      begin
        latest := regpiece(flags);
        if latest = nil then
          begin
            regbranch := nil; exit;
          end;
        flagp := flagp or (flags and HASWIDTH);
        if chain = nil then  { First piece. }
          flagp := flagp or (flags and SPSTART)
        else
          regtail(chain, latest);
        chain := latest;
      end;
    if chain = nil then { Loop ran zero times. }
      regnode(Chr(NOTHING));

    regbranch := ret;
  end;

{
 - regpiece - something followed by possible [*+?]
 *
 * Note that the branching code sequences used for ? and the general cases
 * of * and + are somewhat optimized:  they use the same NOTHING node as
 * both the endmarker for their branch list and the body of the last branch.
 * It might seem that this node could be dispensed with entirely, but the
 * endmarker role is not redundant.
}
function RegularExpression.regpiece( var flagp : word ) : CharPtr;
  var
    ret, next : CharPtr;
    op : char;
    flags : word;
  begin
    ret := regatom(flags);
    if ret = nil then
      begin
        regpiece := nil; exit;
      end;

    op := regparse^;
    if not ISMULT(op) then
      begin
        flagp := flags;
        regpiece := ret; exit;
      end;

    if ((flags and HASWIDTH) = 0) and (op <> '?') then
      begin
        regerror('*+ operand could be empty');
        regpiece := nil; exit;
      end;
    if op = '+' then flagp := WORST or HASWIDTH
    else flagp := WORST or SPSTART;

    if (op = '*') and ((flags and SIMPLE) <> 0) then
      reginsert(Chr(STAR), ret)
    else
      if op = '*' then
        begin
          { Emit x* as (x&|), where & means "self". }
          reginsert(Chr(BRANCH), ret);           { Either x }
          regoptail(ret, regnode(Chr(BACK)));    { and loop }
          regoptail(ret, ret);              { back }
          regtail(ret, regnode(Chr(BRANCH)));    { or }
          regtail(ret, regnode(Chr(NOTHING)));   { null. }
        end
      else
        if (op = '+') and ((flags and SIMPLE) <> 0) then
          reginsert(Chr(PLUS), ret)
        else
          if op = '+' then
            begin
              { Emit x+ as x(&|), where & means "self". }
              next := regnode(Chr(BRANCH));          { Either }
              regtail(ret, next);
              regtail(regnode(Chr(BACK)), ret);      { loop back }
              regtail(next, regnode(Chr(BRANCH)));   { or }
              regtail(ret, regnode(Chr(NOTHING)));   { null. }
            end
          else
            if op = '?' then
              begin
                { Emit x? as (x|) }
                reginsert(Chr(BRANCH), ret);           { Either x }
                regtail(ret, regnode(Chr(BRANCH)));    { or }
                next := regnode(Chr(NOTHING));         { null. }
                regtail(ret, next);
                regoptail(ret, next);
              end;
    Inc(regparse);
    if ISMULT(regparse^) then
      begin
        regerror('nested *?+');
        regpiece := nil; exit;
      end;

    regpiece := ret;
  end;

{
 - regatom - the lowest level
 *
 * Optimization:  gobbles an entire sequence of ordinary characters so that
 * it can turn them into a single node, which is smaller to store and
 * faster to run.  Backslashed characters are exceptions, each becoming a
 * separate node; the code is simpler that way and it's not worth fixing.
}
function RegularExpression.regatom( var flagp : word ) : CharPtr;
  var
    ret : CharPtr;
    flags : word;
    c, ender : char;
    len, classThe, classend : word;
  begin
    flagp := WORST;   { Tentatively. }

    c := regparse^; Inc(regparse);
    case c of
    '^':
      ret := regnode(Chr(BOL));
    '$':
      ret := regnode(Chr(EOL));
    '.':
      begin
        ret := regnode(Chr(ANY));
        flagp := flagp or (HASWIDTH or SIMPLE);
      end;
    '[':
      begin
        if regparse^ = '^' then
          begin
          { Complement of range. }
            ret := regnode(Chr(ANYBUT));
            Inc(regparse);
          end
        else
          ret := regnode(Chr(ANYOF));
        if (regparse^ = ']') or (regparse^ = '-') then
          begin
            regc(regparse^);
            Inc(regparse);
          end;
        while (regparse^ <> #0) and (regparse^ <> ']') do
          begin
            if regparse^ = '-' then
              begin
                Inc(regparse);
                if (regparse^ = ']') or (regparse^ = #0) then
                  regc('-')
                else
                  begin
		  {$IFDEF StonyBrook}
                    classThe := Ord(CharPtr(Ptr(Seg(regparse^),Ofs(regparse^)-2))^)+1;
		  {$ELSE}
                    classThe := Ord((regparse-2)^)+1;
		  {$ENDIF}
                    classend := Ord(regparse^);
                    if classThe > (classend+1) then
                      begin
                        regerror('invalid [] range');
                        regatom := nil; exit;
                      end;
                    while classThe <= classend do
                      begin
                        regc( Chr(classThe) );
                        Inc(classThe);
                      end;
                    Inc(regparse);
                  end
              end
            else
              begin
                if regparse^ = '\' then
                  begin
                    Inc(regparse);
                    regc(escape(regparse));
                  end
                else
                  begin
                    regc(regparse^);
                    Inc(regparse);
                  end;
              end;
          end;
        regc(#0);
        if regparse^ <> ']' then
          begin
            regerror('unmatched []');
            regatom := nil; exit;
          end;
        Inc(regparse);
        flagp := flagp or (HASWIDTH or SIMPLE);
      end;
    '(':
      begin
        ret := reg(1, flags);
        if ret = nil then
          begin
            regatom := nil; exit;
          end;
        flagp := flagp or (flags and (HASWIDTH or SPSTART));
      end;
    #0, '|', ')':
      begin
        regerror('internal urp');
        regatom := nil; exit;
      end; { Supposed to be caught earlier. }
    '?', '+', '*':
      begin
        regerror('?+* follows nothing');
        regatom := nil; exit;
      end;
    '\':
      begin
        if regparse^ = #0 then
          begin
            regerror('trailing \');
            regatom := nil; exit;
          end;
        ret := regnode(Chr(EXACTLY));
        regc(escape(regparse));
        regc(#0);
        flagp := flagp or (HASWIDTH or SIMPLE);
      end;
    else
      begin
        Dec(regparse);
        len := strcspn(regparse, @META);
        if Integer(len) <= 0 then
          begin
            regerror('internal disaster');
            regatom := nil; exit;
          end;
      {$IFDEF StonyBrook}
        ender := CharPtr(Ptr(Seg(regparse^),Ofs(regparse^)+len))^;
      {$ELSE}
        ender := (regparse+len)^;
      {$ENDIF}
        if (len > 1) and ISMULT(ender) then
          Dec(len);    { Back off clear of ?+* operand. }
        flagp := flagp or HASWIDTH;
        if len = 1 then
          flagp := flagp or SIMPLE;
        ret := regnode(Chr(EXACTLY));
        while (len > 0) do
          begin
            regc(regparse^);
            Inc(regparse);
            Dec(len);
          end;
        regc(#0);
      end;
    end;

    regatom := ret;
  end;

{
 - regnode - emit a node
}
function RegularExpression.regnode( op : char ) : CharPtr;
  var
    ret, p : CharPtr;
  begin
    ret := regcode;
    if ret = @regdummy then
      begin
        Inc(regsize,OPHDR);
        regnode := ret; exit;
      end;

    p := ret;
    p^ := op;
    Inc(p);
    p^ := #0;    { Null "next" pointer. }
    Inc(p);
    p^ := #0;
    Inc(p);
    regcode := p;

    regnode := ret;
  end;

{
 - regnext - dig the "next" pointer out of a node
}
function RegularExpression.regnext( p : CharPtr  ) : CharPtr;
  var
    offset : word;
  begin
    if p = @regdummy then
      begin
        regnext := nil; exit;
      end;

    offset := RegOpPtr(p)^.NEXT;
    if offset = 0 then
      begin
        regnext := nil; exit;
      end;

    if RegOpPtr(p)^.OP = BACK then
    {$IFDEF StonyBrook}
      regnext := Ptr(Seg(p^),Ofs(p^)-offset)
    else
      regnext := Ptr(Seg(p^),Ofs(p^)+offset);
    {$ELSE}
      regnext := p-offset
    else
      regnext := p+offset;
    {$ENDIF}
  end;

{
 - reginsert - insert an operator in front of already-emitted operand
 *
 * Means relocating the operand.
}
procedure RegularExpression.reginsert( op : char; opnd : CharPtr );
  var
    src, dst, place : CharPtr;
  begin
    if regcode = @regdummy then
      begin
        Inc(regsize,OPHDR); exit;
      end;

    src := regcode;
    Inc(regcode,OPHDR);
    dst := regcode;
  {$IFDEF StonyBrook}
    while (PtrToLong(src) > PtrToLong(opnd)) do
  {$ELSE}
    while (src > opnd) do
  {$ENDIF}
      begin
        Dec(dst); Dec(src);
        dst^ := src^;
      end;

    place := opnd;   { Op node, where operand used to be. }
    place^ := op; Inc(place);
    place^ := #0; Inc(place);
    place^ := #0; Inc(place);
  end;

{
 - regtail - set the next-pointer at the end of a node chain
}
procedure RegularExpression.regtail( p, val : CharPtr );
  var
    scan, temp : CharPtr;
    offset : word;
  begin
    if p = @regdummy then
      exit;

    { Find last node. }
    scan := p;
    while true do
      begin
        temp := regnext(scan);
        if temp = nil then
          break;
        scan := temp;
      end;

  {$IFDEF StonyBrook}
    offset := PtrDiff(scan,val);
  {$ELSE}
    if RegOpPtr(scan)^.OP = BACK then
      offset := scan - val
    else
      offset := val - scan;
  {$ENDIF}

    RegOpPtr(scan)^.NEXT := offset;
  end;

{
 - regoptail - regtail on operand of first argument; nop if operandless
}
procedure RegularExpression.regoptail( p, val : CharPtr );
  begin
    { "Operandless" and "op != BRANCH" are synonymous in practice. }
    if (p = nil) or (p = @regdummy) or (RegOpPtr(p)^.OP <> BRANCH) then
      exit;
  {$IFDEF StonyBrook}
    regtail(OPERAND(p), val);
  {$ELSE}
    regtail(p + OPHDR, val);
  {$ENDIF}
  end;

{
 - regexec - match a regexp against a string
}
function RegularExpression.regexec( prog : RegExprPtr; st : CharPtr ) : boolean;
  var
    s : CharPtr;
  begin
    { Be paranoid... }
    if (prog = nil) or (st = nil) then
      begin
        regerror('NULL parameter');
        regexec := false; exit;
      end;

    { Check validity of program. }
    if prog^.rprogram[0] <> MAGIC then
      begin
        regerror('corrupted program');
        regexec := false; exit;
      end;

    if regcase then
      begin
        FindChar := StrScan;
        CompString := StrLComp;
      end
    else
      begin
        FindChar := StrIScan;
        CompString := StrLIComp;
      end;
    { If there is a "must appear" string, look for it. }
    if prog^.regmust <> nil then
      begin
        s := st;
        s := FindChar(s, prog^.regmust^);
        while s <> nil do
          begin
            if CompString(s, prog^.regmust, prog^.regmlen) = 0 then
              break;  { Found it. }
            Inc(s);
            s := FindChar(s, prog^.regmust^);
          end;
        if s = nil then { Not present. }
          begin
            regexec := false; exit;
          end;
      end;

    { Mark beginning of line for ^ . }
    regbol := st;

    { Simplest case:  anchored match need be tried only once. }
    if Ord(prog^.reganch) <> 0 then
      begin
        regexec := regtry( prog, st ); exit;
      end;

    { Messy cases:  unanchored match. }
    s := st;
    if Ord(prog^.regstart) <> 0 then
      { We know what char it must start with. }
      begin
        s := FindChar(s, prog^.regstart);
        while s <> nil do
          begin
            if regtry(prog, s) then
              begin
                regexec := true; exit;
              end;
            Inc(s);
            s := FindChar(s, prog^.regstart);
          end
      end
    else
      { We don't -- general case. }
      begin
        Dec(s);
        repeat
          Inc(s);
          if regtry(prog, s) then
            begin
              regexec := true; exit;
            end;
        until s^ = #0;
      end;

    { Failure. }
    regexec := false;
  end;

{
 - regtry - try match at specific point
}
function RegularExpression.regtry( prog : RegExprPtr; st : CharPtr ) : boolean;
  var
    i : word;
    sp, ep : StringTablePtr;
  begin
    reginput := st;
    regstartp := @prog^.startp;
    regendp := @prog^.endp;

    sp := @prog^.startp;
    ep := @prog^.endp;
    for i := 1 to MAX_NSUBEXP do
      begin
        sp^[i] := nil;
        ep^[i] := nil;
      end;
  {$IFDEF StonyBrook}
    if regmatch(CharPtr(Ptr(Seg(prog^.rprogram), Ofs(prog^.rprogram)+ 1))) then
  {$ELSE}
    if regmatch(CharPtr(@prog^.rprogram) + 1) then
  {$ENDIF}
      begin
        prog^.startp[0] := st;
        prog^.endp[0] := reginput;
        regtry := true;
      end
    else
      regtry := false;
  end;

{
 - regmatch - main matching routine
 *
 * Conceptually the strategy is simple:  check to see whether the current
 * node matches, call self recursively to see whether the rest matches,
 * and then act accordingly.  In practice we make some effort to avoid
 * recursion, in particular by going through "ordinary" nodes (that don't
 * need to know whether the rest of the match failed) by a loop instead of
 * by recursion.
}

function RegularExpression.regmatch( prog : CharPtr ) : boolean;
  var
    scan, next, opnd, save : CharPtr;
    len, no, min : word;
    nextch : char;
  begin
    scan := prog;
  {$ifdef DEBUG}
    if (scan != NULL && regnarrate)
      fprintf(stderr, "%s(\n", regprop(scan));
  {$endif}
    while scan <> nil do
      begin
  {$ifdef DEBUG}
      if (regnarrate)
        fprintf(stderr, "%s...\n", regprop(scan));
  {$endif}
        next := regnext(scan);

        case RegOpPtr(scan)^.OP of
        BOL:
          if reginput <> regbol then
            begin
              regmatch := false; exit;
            end;
        EOL:
          if reginput^ <> #0 then
            begin
              regmatch := false; exit;
            end;
        ANY:
          begin
            if reginput^ = #0 then
              begin
                regmatch := false; exit;
              end;
            Inc(reginput);
          end;
        EXACTLY:
          begin
	  {$IFDEF StonyBrook}
            opnd := OPERAND(scan);
	  {$ELSE}
            opnd := scan + OPHDR;
	  {$ENDIF}
            { Inline the first character, for speed. }
            if (regcase and (opnd^ <> reginput^)) or
               ((not regcase) and (UpCase(opnd^) <> UpCase(reginput^))) then
              begin
                regmatch := false; exit;
              end;
            len := StrLen(opnd);
            if (len > 1 ) and (CompString(opnd, reginput, len) <> 0) then
              begin
                regmatch := false; exit;
              end;
            Inc(reginput,len);
          end;
        ANYOF:
          begin
	  {$IFDEF StonyBrook}
            if (reginput^ = #0) or (StrScan(OPERAND(scan), reginput^) = nil) then
	  {$ELSE}
            if (reginput^ = #0) or (StrScan(scan + OPHDR, reginput^) = nil) then
	  {$ENDIF}
              begin
                regmatch := false; exit;
              end;
            Inc(reginput);
          end;
        ANYBUT:
          begin
	  {$IFDEF StonyBrook}
            if (reginput^ = #0) or (StrScan(OPERAND(scan), reginput^) <> nil) then
	  {$ELSE}
            if (reginput^ = #0) or (StrScan(scan+OPHDR, reginput^) <> nil) then
	  {$ENDIF}
              begin
                regmatch := false; exit;
              end;
            Inc(reginput);
          end;
        NOTHING:
          ;
        BACK:
          ;
        OPEN+1,
        OPEN+2,
        OPEN+3,
        OPEN+4,
        OPEN+5,
        OPEN+6,
        OPEN+7,
        OPEN+8,
        OPEN+9:
          begin
            no := RegOpPtr(scan)^.OP - OPEN;
            save := reginput;

            if regmatch(next) then
              begin
                {
                 * Don't set startp if some later
                 * invocation of the same parentheses
                 * already has.
                }
                if regstartp^[no] = nil then
                  regstartp^[no] := save;
                regmatch := true; exit;
              end
            else
              begin
                regmatch := false; exit;
              end;
          end;

        CLOSE+1,
        CLOSE+2,
        CLOSE+3,
        CLOSE+4,
        CLOSE+5,
        CLOSE+6,
        CLOSE+7,
        CLOSE+8,
        CLOSE+9:
          begin
            no := RegOpPtr(scan)^.OP - CLOSE;
            save := reginput;

            if regmatch(next) then
              begin
                {
                 * Don't set endp if some later
                 * invocation of the same parentheses
                 * already has.
                }
                if regendp^[no] = nil then
                  regendp^[no] := save;
                regmatch := true; exit;
              end
            else
              begin
                regmatch := false; exit;
              end;
          end;

        BRANCH:
          begin
            if RegOpPtr(next)^.OP <> BRANCH then   { No choice. }
	    {$IFDEF StonyBrook}
              next := OPERAND(scan) { Avoid recursion. }
	    {$ELSE}
              next := scan + OPHDR { Avoid recursion. }
	    {$ENDIF}
            else
              begin
                repeat
                  save := reginput;
		{$IFDEF StonyBrook}
                  if regmatch(OPERAND(scan)) then
		{$ELSE}
                  if regmatch(scan+OPHDR) then
		{$ENDIF}
                    begin
                      regmatch := true; exit;
                    end;
                  reginput := save;
                  scan := regnext(scan);
                until (scan = nil) or (RegOpPtr(scan)^.OP <> BRANCH);
                regmatch := false; exit;
                { NOTREACHED }
              end;
          end;
        STAR,
        PLUS:
          begin
            {
             * Lookahead to avoid useless match attempts
             * when we know what character comes next.
            }
            nextch := Chr(0);
            if RegOpPtr(next)^.OP = EXACTLY then
	    {$IFDEF StonyBrook}
              nextch := OPERAND(next)^;
	    {$ELSE}
              nextch := (next+OPHDR)^;
	    {$ENDIF}
            if RegOpPtr(scan)^.OP = STAR then
              min := 0
            else min := 1;
            save := reginput;
	  {$IFDEF StonyBrook}
            no := regrepeat(OPERAND(scan));
	  {$ELSE}
            no := regrepeat(scan+OPHDR);
	  {$ENDIF}
            while no >= min do
              begin
                { If it could work, try it. }
                if (nextch = #0) or (reginput^ = nextch) then
                  if regmatch(next) then
                    begin
                      regmatch := true; exit;
                    end;
                { Couldn't or didn't -- back up. }
                Dec(no);
                reginput := save; Inc(reginput,no);
              end;
            regmatch := false; exit;
          end;

        PEND:
          begin
            regmatch := true; exit; { Success! }
          end;

        else
          begin
            regerror('memory corruption');
            regmatch := false; exit;
          end;
        end; {case}

        scan := next;
      end; {while}

    {
     * We get here only if there's trouble -- normally "case END" is
     * the terminating point.
    }
    regerror('corrupted pointers');
    regmatch := false;
  end;

{
 - regrepeat - repeatedly match something simple, report how many
}
function RegularExpression.regrepeat( p : CharPtr ) : word;
  var
    count : word;
    scan, opnd : CharPtr;
  begin
    count := 0;

    scan := reginput;
  {$IFDEF StonyBrook}
    opnd := OPERAND(p);
  {$ELSE}
    opnd := p+OPHDR;
  {$ENDIF}
    case RegOpPtr(p)^.OP of
    ANY:
      begin
        count := StrLen(scan);
        Inc(scan,count);
      end;
    EXACTLY:
      while (opnd^ = scan^) or
            ((not regcase) and (UpCase(opnd^) = UpCase(scan^))) do
        begin
          Inc(count);
          Inc(scan);
        end;
    ANYOF:
      while (scan^ <> #0) and (StrScan(opnd, scan^) <> nil) do
        begin
          Inc(count);
          Inc(scan);
        end;
    ANYBUT:
      while (scan^ <> #0) and (StrScan(opnd, scan^) = nil) do
        begin
          Inc(count);
          Inc(scan);
        end;
    else    { Oh dear.  Called inappropriately. }
      begin
        regerror('internal foulup');
        count := 0;  { Best compromise. }
      end;
    end;
    reginput := scan;

    regrepeat := count;
  end;

function RegularExpression.getRegErrorMsg : string;
  begin
    getRegErrorMsg := regErrorMsg^;
  end;

function RegularExpression.isValid : boolean;
  begin
    isValid := mask <> nil;
  end;

function RegularExpression.maskMatch( addr : CharPtr; caseSense : boolean ) : boolean;
  begin
    regcase := caseSense;
    if mask = nil then
      maskMatch := false
    else maskMatch := regexec( mask, addr );
  end;

{
 - regsub - perform substitutions after a regexp match
}
procedure RegularExpression.regsub( prog : RegExprPtr; source, dest : CharPtr );
  var
    src, dst : CharPtr;
    c : char;
    no : integer;
    len : word;
  begin
    if (prog = nil) or (source = nil) or (dest = nil) then
      begin
        regerror('NULL parm to regsub');
        exit;
      end;
    if prog^.rprogram[0] <> MAGIC then
      begin
        regerror('damaged regexp fed to regsub');
        exit;
      end;

    src := source;
    dst := dest;

    c := src^; Inc(src);
    while ( c <> Chr(0) ) do
      begin
        if c = '&' then
          no := 0
        else if (c = '\') and (Ord('0') <= Ord(src^)) and (Ord(src^) <= Ord('9')) then
          begin
            no := Ord(src^) - Ord('0') + 1;
            Inc(src);
          end
        else
          no := -1;
        if no < 0 then
          begin
            { Ordinary character. }
            if (c = '\') then c := escape(src);
            dst^ := c; Inc(dst);
          end
        else
          if (prog^.startp[no] <> nil) and (prog^.endp[no] <> nil) then
            begin
	    {$IFDEF StonyBrook}
              len := PtrDiff(prog^.endp[no],prog^.startp[no]);
	    {$ELSE}
              len := prog^.endp[no] - prog^.startp[no];
	    {$ENDIF}
              StrLCopy(dst, prog^.startp[no], len);
              Inc(dst,len);
	    {$IFDEF StonyBrook}
              if (len <> 0) and (CharPtr(Ptr(Seg(dst^),Ofs(dst^)-1))^ = Chr(0)) then
	    {$ELSE}
              if (len <> 0) and ((dst-1)^ = Chr(0)) then
	    {$ENDIF}
                begin
                  { strncpy hit NUL. }
                  regerror('damaged match string');
                  exit;
                end;
            end;
        c := src^; Inc(src);
      end;
    dst^ := Chr(0);
  end;

procedure RegularExpression.setReplaceMask( st : CharPtr );
  begin
    if regreplace <> nil then StrDispose( regreplace );
    regreplace := StrNew( st );
  end;

function RegularExpression.maskReplace( addr : CharPtr; caseSense : boolean ) : boolean;
  var
    A : array[0..MaxAsciiz] of char;
    len : word;
    p : CharPtr;
  begin
    regcase := caseSense;
    if (mask = nil) or (regreplace = nil) then
      maskReplace := false
    else
      if regexec( mask, addr ) then
        begin
          p := @A;
	{$IFDEF StonyBrook}
          len := PtrDiff(mask^.startp[0], addr);
	{$ELSE}
          len := mask^.startp[0] - addr;
	{$ENDIF}
          if len <> 0 then
            begin
              StrLCopy( p, addr, len );
              Inc(p, len);
            end;
          regsub( mask, regreplace ,p );
          StrCat( @A, mask^.endp[0] );
          StrCopy( addr, @A );
          maskReplace := true;
        end
      else
        maskReplace := false;
  end;

{$IFOPT D+}
{
 - regdump - dump a regexp onto stdout in vaguely comprehensible form
}
procedure RegularExpression.regdump( r : RegExprPtr );
  var
    s, next : CharPtr;
    op : byte;
  begin
    op := EXACTLY; { Arbitrary non-END op. }


    s := @r^.rprogram;
    Inc(s);
    while op <> PEND do
      begin
        { While that wasn't END last time... }
        op := RegOpPtr(s)^.OP;
        Write( Integer(s - @r^.rprogram), regprop(s)); { Where, what. }
        next := regnext(s);
        if next = nil then  { Next ptr. }
          Write('(0)')
        else
          Write('(',Integer( s - @r^.rprogram ) + Integer(next-s),')');
        Inc(s,OPHDR);
        if (op = ANYOF) or (op = ANYBUT) or (op = EXACTLY) then
          begin
            { Literal string, where present. }
            while s^ <> Chr(0) do
              begin
                Write(s^);
                Inc(s);
              end;
            Inc(s);
          end;
        Writeln;
      end;

    { Header fields of interest. }
    if r^.regstart <> Chr(0) then
      Write('start "', r^.regstart,'"');
    if r^.reganch <> Chr(0) then
      Write('anchored ');
    if r^.regmust <> nil then
      Write('must have "', StrPas(r^.regmust),'"');
    Writeln;
  end;

{
 - regprop - printable representation of opcode
}
function RegularExpression.regprop( op : CharPtr ) : string;
  var
    buf, p : string;

  begin
    buf := ':';

    case RegOpPtr(op)^.OP of
    BOL:
      p := 'BOL';
    EOL:
      p := 'EOL';
    ANY:
      p := 'ANY';
    ANYOF:
      p := 'ANYOF';
    ANYBUT:
      p := 'ANYBUT';
    BRANCH:
      p := 'BRANCH';
    EXACTLY:
      p := 'EXACTLY';
    NOTHING:
      p := 'NOTHING';
    BACK:
      p := 'BACK';
    PEND:
      p := 'END';
    OPEN+1,
    OPEN+2,
    OPEN+3,
    OPEN+4,
    OPEN+5,
    OPEN+6,
    OPEN+7,
    OPEN+8,
    OPEN+9:
      begin
        Str(RegOpPtr(op)^.OP-OPEN,p);
        buf := buf + 'OPEN' + p;
        p := '';
      end;
    CLOSE+1,
    CLOSE+2,
    CLOSE+3,
    CLOSE+4,
    CLOSE+5,
    CLOSE+6,
    CLOSE+7,
    CLOSE+8,
    CLOSE+9:
      begin
        Str(RegOpPtr(op)^.OP-CLOSE,p);
        buf := buf + 'CLOSE' + p;
        p := '';
      end;
    STAR:
      p := 'STAR';
    PLUS:
      p := 'PLUS';
    else
      regerror('corrupted opcode');
    end;
    if p <> '' then
      buf := buf + p;
    regprop := buf;
  end;
{$ENDIF}

end.
